"""Console script for ssl_certinfo."""
import argparse
import ipaddress
import logging
import re
import sys

from ssl_certinfo import __author__, __email__, __version__, ssl_certinfo, validation
from ssl_certinfo.ssl_certinfo import OutputFormat

VERSION = rf"""
ssl_certinfo {__version__}
Copyright (C) 2020 {__author__} ({__email__})
License Apache-2.0: <http://www.apache.org/licenses/LICENSE-2.0>.
"""


def check_hostname_or_ip_address(value):
    """Validate argparse type hostname/ip address."""
    if (
        not validation.is_valid_hostname(value)
        and not validation.is_valid_ip_address(value)
        and not validation.is_valid_ip_network(value)
        and not validation.is_valid_ip_range(value)
    ):
        raise argparse.ArgumentTypeError(
            "%s is not a valid hostname or ip address" % value
        )
    return value


def check_positive(value):
    """Validate argparse type positive integer."""
    try:
        ivalue = int(value)
    except ValueError:
        raise argparse.ArgumentTypeError("%s is not an int value" % value)

    if ivalue <= 0:
        raise argparse.ArgumentTypeError("%s is not a positive int value" % value)
    return ivalue


def check_valid_port(value):
    """Validate argparse type TCP port number."""
    try:
        ivalue = int(value)
    except ValueError:
        raise argparse.ArgumentTypeError("%s is an invalid port number" % value)

    if ivalue <= 0 or ivalue > 65535:
        raise argparse.ArgumentTypeError("%s is an invalid port number" % value)
    return ivalue


def expand_hosts(hostlist):
    result = []

    for elem in hostlist:
        if validation.is_valid_hostname(elem) or validation.is_valid_ip_address(elem):
            result.append(elem)
        elif validation.is_valid_ip_range(elem):
            separator = re.compile(r" *- *")
            (start, end) = separator.split(elem)
            try:
                start_addr = ipaddress.ip_address(start)
                end_addr = ipaddress.ip_address(end)
            except ValueError:
                pass
            else:
                logging.debug("Expanding ip address range " + elem)
                current_ipaddr = start_addr
                while current_ipaddr <= end_addr:
                    result.append(str(current_ipaddr))
                    current_ipaddr = current_ipaddr + 1
        else:
            try:
                net = ipaddress.ip_network(elem, False)
            except ValueError:
                pass
            else:
                logging.debug("Expanding ip network " + elem)
                for ipaddr in net:
                    result.append(str(ipaddr))

    return result


def create_parser():
    """Create ArgParser."""
    parser = argparse.ArgumentParser(
        description="Collect information about SSL certificates from a set of hosts"
    )

    parser.add_argument(
        "-V",
        "--version",
        action="store_true",
        dest="displayVersion",
        help="display version information and exit",
    )

    verb_group = parser.add_mutually_exclusive_group()
    verb_group.add_argument(
        "-v",
        "--verbose",
        action="count",
        dest="verbosity",
        default=0,
        help="verbose output (repeat for increased verbosity)",
    )
    verb_group.add_argument(
        "-q",
        "--quiet",
        action="store_const",
        const=-1,
        default=0,
        dest="verbosity",
        help="quiet output (show errors only)",
    )

    parser.add_argument(
        "host", nargs="*", type=check_hostname_or_ip_address, help="Connect to HOST",
    )

    parser.add_argument(
        "-p",
        "--port",
        default=443,
        type=check_valid_port,
        help="TCP port to connnect to [0-65535]",
    )

    parser.add_argument(
        "-t",
        "--timeout",
        default=5,
        type=check_positive,
        help="Maximum time allowed for connection",
    )

    output_format = parser.add_mutually_exclusive_group()
    output_format.add_argument(
        "-T",
        "--table",
        action="store_const",
        const=OutputFormat.TABLE,
        default=OutputFormat.TABLE,
        dest="outform",
        help="Print results in table format",
    )
    output_format.add_argument(
        "-j",
        "--json",
        action="store_const",
        const=OutputFormat.JSON,
        default=OutputFormat.TABLE,
        dest="outform",
        help="Print results in JSON format",
    )
    output_format.add_argument(
        "-y",
        "--yaml",
        action="store_const",
        const=OutputFormat.YAML,
        default=OutputFormat.TABLE,
        dest="outform",
        help="Print results in YAML format",
    )
    output_format.add_argument(
        "-c",
        "--csv",
        action="store_const",
        const=OutputFormat.CSV,
        default=OutputFormat.TABLE,
        dest="outform",
        help="Print results in CSV format",
    )
    output_format.add_argument(
        "-r",
        "--raw",
        action="store_const",
        const=OutputFormat.RAW,
        default=OutputFormat.TABLE,
        dest="outform",
        help="Print results in raw format",
    )

    return parser


def setup_logging(verbosity):
    base_loglevel = 30
    verbosity = min(verbosity, 2)
    loglevel = base_loglevel - (verbosity * 10)
    logging.basicConfig(level=loglevel, format="%(levelname)s\t%(message)s")


def main():
    """Console script for ssl_certinfo."""
    args = create_parser().parse_args()
    if args.displayVersion:
        print(VERSION)
        return 0

    setup_logging(args.verbosity)

    logging.info("Arguments: " + str(args))

    ssl_certinfo.process_hosts(
        expand_hosts(args.host), args.port, args.timeout, args.outform
    )
    return 0


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



"""Validate ip address and hostname."""
import ipaddress
import re


def is_valid_ip_address(value):
    """Validate if parameter is a valid ip address."""
    try:
        ipaddress.ip_address(value)
        return True
    except ValueError:
        return False


def is_valid_ip_network(value):
    """Validate if parameter is a valid ip network."""
    try:
        net = ipaddress.ip_network(value, False)
    except ValueError:
        return False
    else:
        if net.prefixlen < 32:
            return True
        else:
            return False


def is_valid_ip_range(value):
    """Validate if parameter is a valid ip address range."""
    separator = re.compile(r" *- *")
    try:
        (start, end) = separator.split(value)
        start_addr = ipaddress.ip_address(start)
        end_addr = ipaddress.ip_address(end)
    except ValueError:
        return False
    else:
        if start_addr < end_addr:
            return True
        else:
            return False


def is_valid_hostname(hostname):
    """Validate if parameter is a valid fqdn as defined in RFC-1035."""
    if hostname[-1] == ".":
        # strip exactly one dot from the right, if present
        hostname = hostname[:-1]
    if len(hostname) > 253:
        return False

    labels = hostname.split(".")

    # the TLD must be not all-numeric
    if re.match(r"[0-9]+$", labels[-1]):
        return False

    allowed = re.compile("^[a-z0-9]([a-z0-9-]{0,61}[a-z0-9])?$", re.IGNORECASE)
    return all(allowed.match(label) for label in labels)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ssl_certinfo']

package_data = \
{'': ['*']}

install_requires = \
['pandas', 'pyopenssl', 'pyyaml', 'tabulate', 'tqdm']

entry_points = \
{'console_scripts': ['ssl_certinfo = ssl_certinfo.cli:main']}

setup_kwargs = {
    'name': 'ssl-certinfo',
    'version': '1.0.0',
    'description': 'SSL CertInfo collects information about SSL certificates from a set of hosts.',
    'long_description': '============\nSSL CertInfo\n============\n\n.. start-badges\n\n.. list-table::\n    :stub-columns: 1\n\n    * - build\n      - |travis|\n    * - quality\n      - |codacy| |codeclimate| |sonar-qg| |sonar-rel|\n    * - coverage\n      - |coveralls| |codecov| |codeclimate-cov|\n    * - dependencies\n      - |pyup| |pyup-p3| |requires|\n    * - package\n      - |version| |pyversions| |downloads|\n\n\n.. |travis| image:: https://img.shields.io/travis/stdtom/ssl_certinfo/master.svg?logo=travis\n   :target: https://travis-ci.com/stdtom/ssl_certinfo\n   :alt: Travis Build Status\n\n.. |codacy| image:: https://api.codacy.com/project/badge/Grade/589c03a215ec4ddbb0085b523a857e55\n   :target: https://www.codacy.com/manual/stdtom/ssl_certinfo\n   :alt: Codacy Grade\n\n.. |codeclimate| image:: https://api.codeclimate.com/v1/badges/1ed86e874b3c68672c5c/maintainability\n   :target: https://codeclimate.com/github/stdtom/ssl_certinfo/maintainability\n   :alt: Code Climate Maintainability\n\n.. |sonar-qg| image:: https://sonarcloud.io/api/project_badges/measure?project=stdtom_ssl_certinfo&metric=alert_status\n   :target: https://sonarcloud.io/dashboard?id=stdtom_ssl_certinfo\n   :alt: Sonar Quality Gate Status\n\n.. |sonar-rel| image:: https://sonarcloud.io/api/project_badges/measure?project=stdtom_ssl_certinfo&metric=reliability_rating\n   :target: https://sonarcloud.io/dashboard?id=stdtom_ssl_certinfo\n   :alt: Sonar Reliability Rating\n\n.. |coveralls| image:: https://coveralls.io/repos/github/stdtom/ssl_certinfo/badge.svg?branch=master\n   :target: https://coveralls.io/github/stdtom/ssl_certinfo?branch=master\n   :alt: Coveralls Test Coverage\n\n.. |codecov| image:: https://codecov.io/gh/stdtom/ssl_certinfo/branch/master/graph/badge.svg\n   :target: https://codecov.io/gh/stdtom/ssl_certinfo\n   :alt: CodeCov\n\n.. |codeclimate-cov| image:: https://api.codeclimate.com/v1/badges/1ed86e874b3c68672c5c/test_coverage\n   :target: https://codeclimate.com/github/stdtom/ssl_certinfo/test_coverage\n   :alt: Code Climate Test Coverage\n\n.. |pyup| image:: https://pyup.io/repos/github/stdtom/ssl_certinfo/shield.svg\n   :target: https://pyup.io/repos/github/stdtom/ssl_certinfo/\n   :alt: Updates\n\n.. |pyup-p3| image:: https://pyup.io/repos/github/stdtom/ssl_certinfo/python-3-shield.svg\n   :target: https://pyup.io/repos/github/stdtom/ssl_certinfo/\n   :alt: Python 3\n\n.. |requires| image:: https://requires.io/github/stdtom/ssl_certinfo/requirements.svg?branch=master\n   :target: https://requires.io/github/stdtom/ssl_certinfo/requirements/?branch=master\n   :alt: Requirements Status\n\n.. |version| image:: https://img.shields.io/pypi/v/ssl-certinfo.svg\n   :target: https://pypi.org/project/ssl-certinfo/\n   :alt: Version\n\n.. |pyversions| image:: https://img.shields.io/pypi/pyversions/ssl-certinfo.svg?logo=python&logoColor=FBE072\n    :target: https://pypi.org/project/ssl-certinfo/\n    :alt: Python versions supported\n\n.. |downloads| image:: https://pepy.tech/badge/ssl-certinfo\n    :target: https://pepy.tech/project/ssl-certinfo\n    :alt: PyPI downloads\n\n.. end-badges\n\n\nSSL CertInfo collects information about SSL certificates from a set of hosts.\n\n\nFeatures\n--------\n\n* Hosts to be scanned can be specified as a list of\n\n  * hostnames (fully qualified domain names), e.g. ``github.com``,\n  * ip addresses, e.g. ``1.1.1.1``,\n  * ip networks in CIDR format, e.g. ``10.0.0.0/24``,\n  * ip ranges, e.g. ``10.0.0.1-10.0.0.10``,\n  * or any combination of the previous.\n\n* Results will be presented in various output formats: ``--table``, ``--json``, ``--yaml``, ``--csv``, ``--raw``.\n\n\nInstallation\n------------\nYou can download and install the latest version of this software from the Python package index (PyPI) as follows::\n\n  $ pip install ssl_certinfo\n\n\nUsage\n-----\n\nWhen you install ssl_certinfo, a command-line script called ``ssl_certinfo`` is\nplaced on your path. You can invoke ssl_certinfo directly via this script from the command line::\n\n  $ ssl_certinfo [...]\n\n\nYou can also invoke it through the Python interpreter from the command line::\n\n  $ python -m ssl_certinfo [...]\n\n\nHelp is available with the ``--help`` or ``-h`` switch::\n\n  $ ssl_certinfo -h\n  usage: ssl_certinfo [-h] [-V] [-v | -q] [-p PORT] [-t TIMEOUT] [-T | -j | -y | -c | -r] [host [host ...]]\n\n  Collect information about SSL certificates from a set of hosts\n\n  positional arguments:\n  host                  Connect to HOST\n\n  optional arguments:\n  -h, --help            show this help message and exit\n  -V, --version         display version information and exit\n  -v, --verbose         verbose output (repeat for increased verbosity)\n  -q, --quiet           quiet output (show errors only)\n  -p PORT, --port PORT  TCP port to connnect to [0-65535]\n  -t TIMEOUT, --timeout TIMEOUT\n                        Maximum time allowed for connection\n  -T, --table           Print results in table format\n  -j, --json            Print results in JSON format\n  -y, --yaml            Print results in YAML format\n  -c, --csv             Print results in CSV format\n  -r, --raw             Print results in raw format\n\n\n\n\nCredits\n-------\n\nThis package was created with Cookiecutter_ and the `stdtom/cookiecutter-pypackage-pipenv`_ project template, based on `audreyr/cookiecutter-pypackage`_.\n\n.. _Cookiecutter: https://github.com/cookiecutter/cookiecutter\n.. _`stdtom/cookiecutter-pypackage-pipenv`: https://github.com/stdtom/cookiecutter-pypackage-pipenv\n.. _`audreyr/cookiecutter-pypackage`: https://github.com/audreyr/cookiecutter-pypackage\n',
    'author': 'StdTom',
    'author_email': 'stdtom@gmx.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/stdtom/ssl_certinfo',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""Main module."""
import enum
import json
import logging
import time
from datetime import datetime
from socket import socket

import pandas as pd
import yaml
from cryptography import x509
from cryptography.x509.oid import NameOID
from OpenSSL import SSL
from OpenSSL.SSL import WantReadError, WantWriteError
from tabulate import tabulate
from tqdm import tqdm


class OutputFormat(enum.Enum):
    TABLE = 1
    JSON = 2
    YAML = 3
    CSV = 4
    RAW = 5


def get_cert_info(cert):
    """Get all information about SSL certificate ."""
    certinfo = {}

    certinfo["CN"] = cert.subject.get_attributes_for_oid(NameOID.COMMON_NAME)[0].value

    ext = cert.extensions.get_extension_for_class(x509.SubjectAlternativeName)
    san_list = ext.value.get_values_for_type(x509.DNSName)
    certinfo["SAN"] = ";".join(san_list)

    certinfo["valid_from"] = cert.not_valid_before.isoformat()
    certinfo["valid_to"] = cert.not_valid_after.isoformat()

    delta = cert.not_valid_after - datetime.now()
    certinfo["expire_in_days"] = delta.days

    return certinfo


def ssl_handshake_helper(sock_ssl):
    timeout = sock_ssl.gettimeout()
    if timeout is not None:
        start = time.time()
    while True:
        try:
            return sock_ssl.do_handshake()
        except (WantReadError, WantWriteError):
            if start + timeout <= time.time():
                raise TimeoutError


def get_certificate(hostname, port, timeout=5):
    sock = socket()
    sock.settimeout(timeout)
    sock.connect((hostname, port))

    context = SSL.Context(SSL.SSLv23_METHOD)

    context.check_hostname = False
    context.verify_mode = SSL.VERIFY_NONE

    sock_ssl = SSL.Connection(context, sock)
    sock_ssl.set_connect_state()
    sock_ssl.set_tlsext_host_name(hostname.encode())
    ssl_handshake_helper(sock_ssl)
    cert = sock_ssl.get_peer_certificate()

    sock_ssl.close()
    sock.close()

    return cert.to_cryptography()


def result_to_dataframe(result_dict):
    column_names = [
        "CN",
        "SAN",
        "valid_from",
        "valid_to",
        "expire_in_days",
        "peername",
        "peerport",
    ]
    df = pd.DataFrame(result_dict).T.rename_axis("peer", axis=1)
    df = df.reindex(columns=column_names)

    return df


def process_hosts(hosts, default_port, timeout=5, outform=OutputFormat.TABLE):
    results = {}

    progbar = tqdm(hosts)

    for host in progbar:
        progbar.set_description(f"Checking {host}...")
        try:
            logging.info("Trying to fetch certificate for " + host)
            cert = get_certificate(host, default_port, timeout)
        except (OSError, SSL.Error):
            logging.info("Could not fetch certificate for " + host)
        else:
            certinfo = get_cert_info(cert)
            certinfo["peername"] = host
            certinfo["peerport"] = default_port

            results[host] = certinfo
    print(format_results(results, outform))


def format_results(results, outform):
    if results == {}:
        return ""

    elif outform == OutputFormat.TABLE:
        df = result_to_dataframe(results)
        df.index.name = "peer"
        return tabulate(df, headers="keys", tablefmt="pretty")

    elif outform == OutputFormat.JSON:
        return json.dumps(results, indent=4)

    elif outform == OutputFormat.YAML:
        return yaml.dump(results)

    elif outform == OutputFormat.CSV:
        df = result_to_dataframe(results)
        df.index.name = "peer"
        return df.to_csv()

    elif outform == OutputFormat.RAW:
        df = result_to_dataframe(results)
        return df.to_string()
